# -*- coding: utf-8 -*-
"""
Created on Sun Feb 26 16:41:01 2017

@author: andi
"""

import visr
import rbbl
import rcl
import pml
import panning

# Usage in debugger:
# exec(open("./PanningComparison.py").read())

import numpy as np
#import matplotlib.pyplot as plt

from audio_signal_switch import AudioSignalSwitch
from pointsource_trajectory_generator import PointSourceTrajectoryGenerator
from wav_playback import WavPlayback


class PanningComparison( visr.CompositeComponent ):
    def __init__( self, context, name, parent,
                 rendererConfigs,
                 numberOfInputs,
                 numberOfOutputs,
                 interpolationSteps = None,
                 controlDataType = pml.Float,
                 outputChannelRouting = None,
                 delayAdjust = None,
                 gainAdjust = None ):
        super(PanningComparison,self).__init__( context, name, parent )

        if interpolationSteps is None:
            interpolationSteps = context.period

        numberOfRenderers = len( rendererConfigs )

        self.input = visr.AudioInputFloat( "in", self, numberOfInputs )
        self.objectInput = visr.ParameterInput("objectIn", self,
                                               protocolType=pml.DoubleBufferingProtocol.staticType,
                                               parameterType=pml.ObjectVector.staticType,
                                               parameterConfig=pml.EmptyParameterConfig() )
        self.switchInput = visr.ParameterInput( "switch", self,
                                                protocolType=pml.MessageQueueProtocol.staticType,
                                                parameterType=controlDataType.staticType,
                                                parameterConfig=pml.EmptyParameterConfig() )

        self.signalSwitch = AudioSignalSwitch( context, "OutputSwitch", self,
                                         numberOfChannels=numberOfOutputs,
                                         numberOfInputs=numberOfRenderers,
                                         controlDataType=controlDataType
                                         )
        self.parameterConnection( self.switchInput, self.signalSwitch.parameterPort('controlIn') )
        if outputChannelRouting is None:
            outputChannelRouting = list(range(numberOfOutputs) ) # Create a 1:1 routing
        if len( outputChannelRouting ) != numberOfOutputs:
            ValueError( "Optional argument OutputChannelRouting must have the same length as numberOfOtputs." )
        maxChannel = np.max( outputChannelRouting )

        self.output = visr.AudioOutputFloat( "out", self, maxChannel+1 )
        if delayAdjust is None and gainAdjust is None:
            self.audioConnection( self.signalSwitch.audioPort('out'), range(numberOfOutputs),
                                 self.output, list(outputChannelRouting)  )
        else:
            if delayAdjust is None:
                delayAdjustVec = np.zeros( numberOfOutputs, np.float32 )
            else:
                delayAdjustVec = np.asarray( delayAdjust, np.float32 )
            if gainAdjust is None:
                gainAdjustVec = np.zeros( numberOfOutputs, np.float32 )
            else:
                gainAdjustVec = np.asarray( gainAdjust, np.float32 )
            self.outputAdjust = rcl.DelayVector( context, "OutputAdjust", self,
                                                numberOfChannels = numberOfOutputs,
                                                maxDelay = np.max( delayAdjustVec ),
                                                interpolationType = "lagrangeOrder0",
                                                controlInputs = rcl.DelayVector.ControlPortConfig.No,
                                                initialDelay = delayAdjustVec,
                                                initialGain = gainAdjustVec
                                                )
            self.audioConnection( self.signalSwitch.audioPort('out'),
                                 self.outputAdjust.audioPort("in")  )
            self.audioConnection( self.outputAdjust.audioPort('out'), range(numberOfOutputs),
                                 self.output, list(outputChannelRouting)  )



        unusedOutputChannels = set(range(maxChannel)) - set(outputChannelRouting)
        if len(unusedOutputChannels) > 0:
            self.outputSilencer = NullSource( context, "OutputSilencer", self, 1 )
            self.audioConnection( self.outputSilencer.audioPort('out'), [0]*len(unusedOutputChannels),
                                 self.output, list(unusedOutputChannels) )

        self.rendererBank = []
        for rendererIdx, rendererCfg  in enumerate(rendererConfigs):
            rendererClass = rendererCfg["className"]
            rendererName = rendererCfg["name"]
            rendererArgs = rendererCfg["kwArgs"]
            newRenderer = rendererClass( context, rendererName, self,
                                        interpolationSteps=context.period,
                                        numberOfInputs=numberOfInputs, **rendererArgs )
            self.audioConnection( self.input, newRenderer.audioPort("in") )
            self.audioConnection( newRenderer.audioPort("out"), self.signalSwitch.audioPort("in_%d" % rendererIdx ) )
            self.parameterConnection( self.objectInput, newRenderer.parameterPort("objects") )
            self.rendererBank.append( newRenderer )

class AudioTerminator( visr.AtomicComponent ):
    """ Simple component to connect unneeded audio outputs to. """
    def __init__( self, context, name, parent, width ):
        super( AudioTerminator, self ).__init__( context, name, parent )
        self.input = visr.AudioInputFloat( "in", self, width=width )
    def process( self ):
        pass # Nothing to do

class PanningComparisonTrajectory( visr.CompositeComponent ):
    """ Extends a PanningComparison component by a trajectory generator and optional an audio playback"""
    def __init__( self, context, name, parent,
                 rendererConfigs,
                 numberOfInputs,
                 numberOfOutputs,
                 trajectoryPositions,
                 trajectoryUpdateSamples = None,
                 sourceLevel=1.0,
                 sourceObjectId = 0,
                 controlDataType = pml.Float,
                 testSignal = None,
                 useDummyInput = False,
                 outputChannelRouting = None,
                 delayAdjust = None,
                 gainAdjust = None,
                 trajectoryControlInputType = None,  # Whether a parameter input port 'control' is instantiated to switch between
                                                     # different trajectories.
                 initialTrajectoryIndex = None
                 ):

        if trajectoryUpdateSamples is None:
            trajectoryUpdateSamples = context.period

        super(PanningComparisonTrajectory,self).__init__( context, name, parent)
        # Some devices (e.g., ASIO) have problems with 0 inputs, therefore we create a dummy input if needed.
        if testSignal is None or useDummyInput:
            self.input = visr.AudioInputFloat( "in", self, numberOfInputs )

        self.switchInput = visr.ParameterInput( "switch", self,
                                                protocolType=pml.MessageQueueProtocol.staticType,
                                                parameterType=controlDataType.staticType,
                                                parameterConfig=pml.EmptyParameterConfig() )

        self.multiRenderer = PanningComparison(context, "MultiRenderer", self,
                                           rendererConfigs=rendererConfigs,
                                           numberOfInputs=numberOfInputs,
                                           numberOfOutputs=numberOfOutputs,
                                           interpolationSteps=trajectoryUpdateSamples,
                                           controlDataType=controlDataType,
                                           outputChannelRouting = outputChannelRouting,
                                           delayAdjust = delayAdjust,
                                           gainAdjust = gainAdjust,
)
        if testSignal is not None:
            self.signalGenerator = WavPlayback( context, "SignalGenerator", self,
                                               testSignal, numberOfChannels=numberOfInputs)
            self.audioConnection( self.signalGenerator.audioPort( "out" ),
                                 self.multiRenderer.audioPort( "in" )  )
            if useDummyInput:
                self.dummy = AudioTerminator( context, "audioTerminator", self, numberOfInputs )
                self.audioConnection( self.input, self.dummy.audioPort( "in" ) )
        else:
            self.audioConnection( self.input, self.multiRenderer.audioPort( "in" ) )

        outputWidth = self.multiRenderer.audioPort( "out" ).width
        self.output = visr.AudioOutputFloat( "out", self, outputWidth )
        self.audioConnection( self.multiRenderer.audioPort( "out" ), self.output )

        self.sceneGenerator = PointSourceTrajectoryGenerator( context, "SceneDecoder", self,
                 positions=trajectoryPositions,
                 updateRateSamples = trajectoryUpdateSamples,
                 objectId = sourceObjectId,
                 groupId = 0,
                 priority = 0,
                 objectLevel = sourceLevel,
                 controlInputType = trajectoryControlInputType,
                 initialTrajectoryIndex = 0
                 )

        self.parameterConnection( self.sceneGenerator.parameterPort( "objectVectorOutput"),
                                 self.multiRenderer.parameterPort( "objectIn" ) )

        self.parameterConnection( self.switchInput,
                                 self.multiRenderer.parameterPort( "switch" ) )

        if trajectoryControlInputType is not None:
            self.trajectoryControlInput = visr.ParameterInput( "trajectoryControlInput", self,
                                    parameterType=trajectoryControlInputType.staticType,
                                    protocolType=pml.MessageQueueProtocol.staticType,
                                    parameterConfig=pml.EmptyParameterConfig() )
            self.parameterConnection( self.trajectoryControlInput,
                                     self.sceneGenerator.parameterPort("controlInput") )

class RealTimePanningComparison( visr.CompositeComponent ):
    """
    Signal flow
    """

    def __init__( self, context, name, parent,
                 rendererConfigs,
                 numberOfInputs,
                 numberOfOutputs,
                 interpolationSteps,
                 sceneReceivePort = 4242,
                 controlDataType = pml.Float,
                 controlReceivePort=8889,
                 delayAdjust = None,
                 gainAdjust = None,
                 outputchannelRouting = None
                 ):
        super(RealTimePanningComparison,self).__init__( context, name, parent )
        self.renderer = PanningComparison(
                 context, name, self,
                 rendererConfigs = rendererConfigs,
                 numberOfInputs = numberOfInputs,
                 numberOfOutputs = numberOfOutputs,
                 interpolationSteps = interpolationSteps,
                 controlDataType = controlDataType,
                 outputChannelRouting = outputChannelRouting,
                 delayAdjust = delayAdjust,
                 gainAdjust = gainAdjust )
        TypeStrings = { pml.Float: 'float', pml.Double: 'double',
                       pml.Integer: 'int', pml.UnsignedInteger: 'uint'  }
        dtype = TypeStrings[controlDataType]
        self.controlReceiver = rcl.UdpReceiver( context, "ControlReceiver", self,
                                               port=controlReceivePort,
                                               mode=rcl.UdpReceiver.Mode.Asynchronous)
        self.controlDecoder = rcl.ScalarOscDecoder( context, "ControlDecoder", self )
        self.controlDecoder.setup(dataType=dtype)
        self.parameterConnection( self.controlReceiver.parameterPort("messageOutput"),
                                 self.controlDecoder.parameterPort("datagramInput") )
        self.parameterConnection( self.controlDecoder.parameterPort( "dataOut"),
                                 self.renderer.parameterPort( "switch" ) )
        self.sceneReceiver = rcl.UdpReceiver( context, "sceneReceiver", self,
                                               port=sceneReceivePort,
                                               mode=rcl.UdpReceiver.Mode.Asynchronous)
        self.sceneDecoder = rcl.SceneDecoder( context, "SceneDecoder", self )
        self.parameterConnection( self.sceneReceiver.parameterPort("messageOutput"),
                                 self.sceneDecoder.parameterPort("datagramInput") )
        self.parameterConnection( self.sceneDecoder.parameterPort( "objectVectorOutput"),
                                 self.renderer.parameterPort( "objectIn" ) )

        self.input = visr.AudioInputFloat( "in", self, numberOfInputs )
        self.audioConnection( self.input, self.renderer.audioPort("in") )
        outputWidth = self.renderer.audioPort( "out" ).width
        self.output = visr.AudioOutputFloat( "out", self, outputWidth )
        self.audioConnection( self.renderer.audioPort("out"), self.output )

class RealTimePanningComparisonTrajectory( visr.CompositeComponent ):
    """

    """
    def __init__( self, context, name, parent,
                 rendererConfigs,
                 numberOfInputs,
                 numberOfOutputs,
                 trajectoryPositions,
                 trajectoryUpdateSamples = None,
                 sourceLevel=1.0,
                 sourceObjectId = 0,
                 controlDataType = pml.Float,
                 controlReceivePort=8889,
                 testSignal = None,
                 useDummyInput = False,
                 outputChannelRouting = None,
                 delayAdjust = None,
                 gainAdjust = None,
                 trajectoryControlInputType = None,
                 trajectoryControlPortNumber = 9999,
                 initialTrajectoryIndex = 0
                 ):
        super(RealTimePanningComparisonTrajectory,self).__init__( context, name, parent )
        self.renderer = PanningComparisonTrajectory(
                 context, name, self,
                 rendererConfigs = rendererConfigs,
                 numberOfInputs = numberOfInputs,
                 numberOfOutputs = numberOfOutputs,
                 trajectoryPositions = trajectoryPositions,
                 trajectoryUpdateSamples = trajectoryUpdateSamples,
                 sourceLevel = sourceLevel,
                 sourceObjectId = sourceObjectId,
                 controlDataType = controlDataType,
                 testSignal = testSignal,
                 useDummyInput = useDummyInput,
                 outputChannelRouting = outputChannelRouting,
                 delayAdjust = delayAdjust,
                 gainAdjust = gainAdjust,
                 trajectoryControlInputType = trajectoryControlInputType,
                 initialTrajectoryIndex = initialTrajectoryIndex )
        if testSignal is None or useDummyInput:
            self.input = visr.AudioInputFloat( "in", self, numberOfInputs )
            self.audioConnection( self.input, self.renderer.audioPort("in") )
        outputWidth = self.renderer.audioPort( "out" ).width
        self.output = visr.AudioOutputFloat( "out", self, outputWidth )
        self.audioConnection( self.renderer.audioPort("out"), self.output )


        self.controlReceiver = rcl.UdpReceiver( context, "ControlReceiver", self,
                                               port=controlReceivePort,
                                               mode=rcl.UdpReceiver.Mode.Asynchronous)


        TypeStrings = { pml.Float: 'float', pml.Double: 'double',
                       pml.Integer: 'int', pml.UnsignedInteger: 'uint'  }
        dtype = TypeStrings[controlDataType]
        self.controlDecoder = rcl.ScalarOscDecoder( context, "ControlDecoder", self )
        self.controlDecoder.setup(dataType=dtype)
        self.parameterConnection( self.controlReceiver.parameterPort("messageOutput"),
                                 self.controlDecoder.parameterPort("datagramInput") )
        self.parameterConnection( self.controlDecoder.parameterPort( "dataOut"),
                                 self.renderer.parameterPort( "switch" ) )


        if trajectoryControlInputType is not None:
            self.trajectoryControlReceiver =  rcl.UdpReceiver( context, "TrajectoryReceiver", self,
                                               port=trajectoryControlPortNumber,
                                               mode=rcl.UdpReceiver.Mode.Asynchronous)
            dtype = TypeStrings[trajectoryControlInputType]
            self.trajectoryControlDecoder = rcl.ScalarOscDecoder( context, "TrajectoryControlDecoder", self )
            self.trajectoryControlDecoder.setup(dataType=dtype)
            self.parameterConnection( self.trajectoryControlReceiver.parameterPort("messageOutput"),
                                     self.trajectoryControlDecoder.parameterPort("datagramInput") )
            self.parameterConnection( self.trajectoryControlDecoder.parameterPort( "dataOut"),
                                     self.renderer.parameterPort( "trajectoryControlInput" ) )

class NullSource( visr.AtomicComponent ):
    """
    Component computing the element-wise square root of a matrix.
    """
    def __init__( self, context, name, parent,
                 width = 1,
                 ):
        super( NullSource, self ).__init__(context, name, parent )
        self.output = visr.AudioOutputFloat( "out", self, width )
    def process( self ):
        outSig = np.asarray( self.output.data() )
        outSig = 0.0
