# -*- coding: utf-8 -*-
"""
Created on Mon Feb 26 19:39:04 2018

@author: andi
"""

import visr
import pml
import rbbl
import rcl

from vbap_l1_l2_panner import VbapL1L2Panner

from enum import Enum
import numpy as np

class FrequencyType( Enum ):
    Low = 1
    High = 2,
    Dual = 3

class VbapRenderer( visr.CompositeComponent ):
    def __init__(self, context, name, parent,
             numberOfInputs,
             arrayConfig,
             interpolationSteps,
             bandType = FrequencyType.Low,
             normalisationNorm = None,   # Default value
             l2Panner = False,
             filterBank = None
             ):
        super( VbapRenderer, self ).__init__(context, name, parent )
        numberOfOutputs = arrayConfig.numberOfRegularLoudspeakers
        self.input = visr.AudioInputFloat( "in", self, numberOfInputs )
        self.output = visr.AudioOutputFloat( "out", self, numberOfOutputs )

        self.objectInput = visr.ParameterInput( "objects", self,
                                               parameterType=pml.ObjectVector.staticType,
                                               protocolType=pml.DoubleBufferingProtocol.staticType,
                                               parameterConfig=pml.EmptyParameterConfig() )

        if l2Panner:
            self.panningCalculator = VbapL1L2Panner( context, "GainCalculator", self,
                                                    lspArray=arrayConfig,
                                                    numObjects=numberOfInputs )
        else:
            self.panningCalculator = rcl.PanningCalculator( context, "GainCalculator", self,
                                                           arrayConfig=arrayConfig,
                                                           numberOfObjects=numberOfInputs )
        self.parameterConnection( self.objectInput,
                                 self.panningCalculator.parameterPort( "objectVectorInput" ) )


        if bandType in [FrequencyType.Low, FrequencyType.Dual]:
            if normalisationNorm is None:
                lfNorm = 1
            elif np.size(normalisationNorm) == 1:
                lfNorm = normalisationNorm
            else:
                lfNorm = normalisationNorm[0]
            self.normaliserLow = Normaliser( context, "NormaliserLow", self,
                                            numberOfOutputs, numberOfInputs, lfNorm )
            self.parameterConnection( self.panningCalculator.parameterPort("gainOutput"),
                                     self.normaliserLow.parameterPort("in") )
            self.gainMatrixLow = rcl.GainMatrix( context, "PanningMatrixLow", self,
                                             numberOfInputs=numberOfInputs,
                                             numberOfOutputs=numberOfOutputs,
                                             interpolationSteps=interpolationSteps,
                                             initialGains = 0.0,
                                             controlInput=True )
            self.parameterConnection( self.normaliserLow.parameterPort("out"),
                                     self.gainMatrixLow.parameterPort("gainInput"))

        if bandType in [FrequencyType.High, FrequencyType.Dual]:
            if normalisationNorm is None:
                hfNorm = 1
            elif np.size(normalisationNorm) == 1:
                hfNorm = normalisationNorm
            else:
                hfNorm = normalisationNorm[1]
            self.squareRootHigh=MatrixSquareRoot( context, "SquareRootHigh", self,
                                                 numberOfOutputs, numberOfInputs )
            self.parameterConnection( self.panningCalculator.parameterPort("gainOutput"),
                                     self.squareRootHigh.parameterPort("in") )
            self.normaliserHigh = Normaliser( context, "NormaliserHigh", self,
                                            numberOfOutputs, numberOfInputs, hfNorm )
            self.parameterConnection( self.squareRootHigh.parameterPort("out"),
                                     self.normaliserHigh.parameterPort("in") )
            self.gainMatrixHigh = rcl.GainMatrix( context, "PanningMatrixHigh", self,
                                             numberOfInputs=numberOfInputs,
                                             numberOfOutputs=numberOfOutputs,
                                             interpolationSteps=interpolationSteps,
                                             initialGains = 0.0,
                                             controlInput=True )
            self.parameterConnection( self.normaliserHigh.parameterPort("out"),
                                     self.gainMatrixHigh.parameterPort("gainInput"))

        if bandType == FrequencyType.Dual:
            if filterBank is None:
                filterBank = [rbbl.BiquadCoefficientFloat( 0.001921697757295, 0.003843395514590,
                              0.001921697757295,-1.824651307057289, 0.832338098086468 ),
                           rbbl.BiquadCoefficientFloat( -0.914247351285939, 1.828494702571878,
                             -0.914247351285939, -1.824651307057289, 0.832338098086468 ) ]
            lfFilter = filterBank[0]
            hfFilter = filterBank[1]
            filterMtx = rbbl.BiquadCoefficientMatrixFloat( 2*numberOfInputs, 1 )
            for idx in range(0,numberOfInputs ):
                filterMtx[idx,0] = lfFilter
                filterMtx[idx+numberOfInputs,0] = hfFilter

            self.filterBank = rcl.BiquadIirFilter( context, "filterBank", self,
                                                  numberOfChannels=2*numberOfInputs,
                                                  numberOfBiquads=1, # TODO: allow variable number of sections.
                                                  initialBiquads=filterMtx,
                                                  controlInput = False )
            self.combiner = rcl.Add( context, "LfHfGain", self,
                                    numInputs = 2,
                                    width=numberOfOutputs)

            self.audioConnection( self.input,
                                 [ i % numberOfInputs for i in range(0, 2*numberOfInputs) ],
                                 self.filterBank.audioPort("in"),
                                 range(0,2*numberOfInputs))
            self.audioConnection( self.combiner.audioPort("out"), self.output )

            self.audioConnection( self.filterBank.audioPort("out"), range(numberOfInputs),
                                 self.gainMatrixLow.audioPort("in"),  range(numberOfInputs) )
            self.audioConnection( self.filterBank.audioPort("out"), range(numberOfInputs,2*numberOfInputs),
                                 self.gainMatrixHigh.audioPort("in"),  range(numberOfInputs) )
            self.audioConnection( self.gainMatrixLow.audioPort("out"), self.combiner.audioPort("in0") )
            self.audioConnection( self.gainMatrixHigh.audioPort("out"), self.combiner.audioPort("in1") )
        elif bandType == FrequencyType.Low:
            self.audioConnection( self.input, self.gainMatrixLow.audioPort("in") )
            self.audioConnection( self.gainMatrixLow.audioPort("out"), self.output )
        elif bandType == FrequencyType.High:
            self.audioConnection( self.input, self.gainMatrixHigh.audioPort("in") )
            self.audioConnection( self.gainMatrixHigh.audioPort("out"), self.output )


class MatrixSquareRoot( visr.AtomicComponent ):
    """
    Component computing the element-wise square root of a matrix.
    """
    def __init__( self, context, name, parent,
                 numberOfRows,
                 numberOfColumns
                 ):
        super( MatrixSquareRoot, self ).__init__(context, name, parent )
        paramCfg = pml.MatrixParameterConfig(numberOfRows, numberOfColumns )
        self.mtxIn = visr.ParameterInput( "in", self,
                                         parameterType=pml.MatrixParameterFloat.staticType,
                                         protocolType=pml.SharedDataProtocol.staticType,
                                         parameterConfig=paramCfg )
        self.mtxOut = visr.ParameterOutput( "out", self,
                                         parameterType=pml.MatrixParameterFloat.staticType,
                                         protocolType=pml.SharedDataProtocol.staticType,
                                         parameterConfig=paramCfg )

    def process( self ):
        inGains = np.asarray( self.mtxIn.protocolInput().data() )
        outGains = np.asarray( self.mtxOut.protocolOutput().data() )
        outGains[...] = np.sqrt( inGains )

class Normaliser( visr.AtomicComponent ):
    """
    Component performing a row-wise normalisation.
    """
    def __init__( self, context, name, parent,
                 numberOfRows,
                 numberOfColumns,
                 lpNorm = 2
                 ):
        super( Normaliser, self ).__init__(context, name, parent )
        self.lpNorm = lpNorm
        paramCfg = pml.MatrixParameterConfig(numberOfRows, numberOfColumns )
        self.mtxIn = visr.ParameterInput( "in", self,
                                         parameterType=pml.MatrixParameterFloat.staticType,
                                         protocolType=pml.SharedDataProtocol.staticType,
                                         parameterConfig=paramCfg )
        self.mtxOut = visr.ParameterOutput( "out", self,
                                         parameterType=pml.MatrixParameterFloat.staticType,
                                         protocolType=pml.SharedDataProtocol.staticType,
                                         parameterConfig=paramCfg )

    def process( self ):
        inGains = np.nan_to_num( self.mtxIn.protocolInput().data() )
        outGains = np.asarray( self.mtxOut.protocolOutput().data() )
        if self.lpNorm == 0:
            outGains[...] = inGains
        else:
            srcNorms = np.linalg.norm( inGains, ord=self.lpNorm, axis = 0 )
            srcNorms[ np.abs( srcNorms) < 1e-8 ] = 1.0 # Avoid divide by zero or values very close to zero.
            outGains[...] = 1.0/srcNorms[np.newaxis,:] * inGains
